import datetime
import logging

from typing import Union

from . import exceptions

logger = logging.getLogger(__name__)

TIMESTAMP_ATTRIBUTES = {
    "accountExpires",
    "badPasswordTime",
    "lastLogon",
    "lastLogonTimestamp",
    "lockoutTime",
    "pwdLastSet",
}


class TranslatedTimestamp:
    def __init__(self, unix_epoch: int, windows_epoch: int):
        self.unix_epoch = unix_epoch
        self.windows_epoch = windows_epoch
        self.date_object = datetime.datetime.utcfromtimestamp(unix_epoch)

    def __str__(self):
        return str(self.date_object)

    def __repr__(self):
        return f"TranslatedTimestamp(unix_date={self.__str__()}, windows_epoch={self.windows_epoch})"

    def __eq__(self, other):
        if isinstance(other, datetime.datetime):
            return self.date_object == other
        return str.__eq__(str(self), str(other))

    def __ne__(self, other):
        return not self.__eq__(other)


def windows_bytes_to_unix_date(raw_timestamp: bytes) -> Union[TranslatedTimestamp, None]:
    # Creates a TranslatedTimestamp instance from an LDAP windows epoch timestamp. The raw
    # epoch arrives as bytes as it won't have been decoded from the LDAP response yet.
    if not isinstance(raw_timestamp, bytes):
        raise exceptions.AttributeDecodingError(
            f"Can only decode windows timestamp from bytes. Received {type(raw_timestamp)}."
        )

    logger.debug(f"converting active directory epoch {raw_timestamp} to datetime")
    try:
        windows_epoch = int(raw_timestamp.decode("utf-8"))
    except (ValueError, AttributeError):
        raise exceptions.AttributeDecodingError(f"Failed to decode windows timestamp with value {raw_timestamp}.")

    if windows_epoch in {9_223_372_036_854_775_807, 0}:  # never expires or not set
        return None

    # Convert AD value (100 nano-second intervals since 1/1/1601) to seconds
    seconds = int(windows_epoch / 10_000_000)

    # Subtract the difference in seconds between AD epoch (1/1/1601) and UNIX epoch (1/1/1970) so we can
    # create a datetime object from a UNIX epoch.
    unix_epoch = int(seconds) - 11_644_473_600

    try:
        return TranslatedTimestamp(unix_epoch, windows_epoch)
    except (ValueError, OverflowError) as e:
        if "out of range" in str(e):
            # A 'never expires' or 'never set' value in AD will be a timestamp in the distant future. In python-land
            # these values should be None. This shouldn't happen because we catch 0 and 9223372036854775807 further
            # up, but better safe than sorry.
            logger.debug(f"UNIX epoch stamp {unix_epoch} is too far in the future to convert.")
            return None
        raise e
