"""
Groups

This module contains actions related to Active Directory Groups.
"""
import uuid
import logging

from typing import List, Dict, Optional, Union, Any


from ppa import ui
from active_directory import helpers, users, groups, computers

from .exceptions import NoGroupFound
from ._wrappers import ldap_function, idempotent_create, idempotent_remove
from .records import Group, GroupData, User, UserData, Computer, ComputerData, DomainController

logger = logging.getLogger(__name__)


@ldap_function
def get_all(domain_controller: DomainController, search_base: Optional[str] = None) -> List[Group]:
    """Get all groups.

!!! danger "This operation can be slow on large domains"

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __search_base:__ a distinguishedName to use as the root of the search (defaults to the root of the domain)

??? success "Output"

    A list of [Group Dictionaries](../inputs_outputs/#group).

??? example
    ```YAML linenums="1"
    - active_directory.groups.get_all:
      load:
        domain_controller: domain_controller_info
      save: all_groups
    ```
"""
    return [GroupData(**group.__dict__).__dict__ for group in groups.get_all(domain_controller, search_base=search_base)]


@ldap_function
def by_samaccountname(domain_controller: DomainController, sAMAccountName: str) -> Group:
    """Get a group by its 'sAMAccountName' attribute.

!!! question "Minimum Plugin Version: 1.3.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __sAMAccountName:__ The group's `sAMAccountName`

??? success "Output"
    A single [Group Dictionary](../inputs_outputs/#group).

??? example
    ```YAML linenums="1"
    - active_directory.groups.by_samaccountname:
        sAMAccountName: example.group
      load:
        domain_controller: domain_controller_info
      save: example_group
    ```
"""
    try:
        return GroupData(**groups.search(domain_controller, sAMAccountName=sAMAccountName)[0].__dict__).__dict__
    except IndexError:
        raise NoGroupFound(f"No group found with sAMAccountName '{sAMAccountName}'.")


@ldap_function
def by_distinguishedname(domain_controller: DomainController, distinguishedName: str) -> Group:
    """Get a group by its 'distinguishedName' attribute.

!!! question "Minimum Plugin Version: 1.3.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ The group's `distinguishedName`

??? success "Output"
    A single [Group Dictionary](../inputs_outputs/#group).

??? example
    ```YAML linenums="1"
    - active_directory.groups.by_distinguishedname:
        distinguishedName: cn=Example Group,CN=Users,DC=Example,DC=Domain
      load:
        domain_controller: domain_controller_info
      save: example_group
    ```
"""
    try:
        return GroupData(**groups.search(domain_controller, distinguishedName=distinguishedName)[0].__dict__).__dict__
    except IndexError:
        raise NoGroupFound(f"No group found with distinguishedName '{distinguishedName}'.")


@ldap_function
def get_interactive(
    domain_controller: DomainController,
    title: Optional[str] = None,
    exclude: Optional[Dict[str, Union[str, int]]] = None,
    search_base: Optional[str] = None,
) -> Group:
    """Start an interactive search for a group using its 'Common Name'.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __title:__ a title displayed to the Task Operator (defaults to `Get Security Group`)

    - __exclude:__ a dictionary containing LDAP attribute names & regular expressions to test their values

    - __search_base:__ a distinguishedName to use as the root of the search (defaults to the root of the domain)

??? success "Output"

    A single [Group Dictionary](../inputs_outputs/#group).

!!! tip "Automatic Wildcards"

    When using this action all provided search terms will have a wildcard appended.

??? example

    - Interactively searching for a group inside the builtin __Users__ CN

    - All groups whose `cn` contains __admin__ are excluded from the results

    ```YAML linenums="1"
    - active_directory.groups.get_interactive:
        search_base: CN=Users,DC=Example,DC=Domain,DC=Com
        exclude:
          cn: .*admin.*
      load:
        domain_controller: domain_controller_info
      save: group
    ```
    """
    return GroupData(
        **helpers.get_group(
            domain_controller,
            search_attribute="cn",
            search_base=search_base,
            auto_wildcard=True,
            auto_select=False,
            exclude=exclude,
            title=title,
        ).__dict__
    ).__dict__


@ldap_function
def get_members(
    domain_controller: DomainController, distinguishedName: str, nested: Optional[bool] = False
) -> List[Union[Group, User, Computer]]:
    """Get the members of a group.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __distinguishedName:__ the distinguished name of the group

    - __nested:__ set to `true` to include nested group members in the search

??? success "Output"
    A list of [User](../inputs_outputs/#user), [Group](../inputs_outputs/#group), & [Computer](../inputs_outputs/#computer) dictionaries.

??? warning
    Getting nested group memberships can be slow.

??? example
    Getting all members of _Example Group_, & saving them as a new variable called _group_members_.
    ```YAML linenums="1"
    - active_directory.groups.get_members:
        distinguishedName: "CN=Example Group,CN=Users,DC=Example,DC=Domain"
        nested: true
      load:
        domain_controller: domain_controller_info
      save: group_members
    ```
    """
    record_generators = {
        "User": UserData,
        "Group": GroupData,
        "Computer": ComputerData,
    }
    try:
        return [
            record_generators[member._ppa_record_name](**member.__dict__).__dict__
            for member in groups.members(domain_controller, distinguished_name=distinguishedName, nested=nested)
        ]
    except KeyError:
        raise KeyError("Group members contains a record that is not one of User, Group, Computer. Cannot continue.")


@ldap_function
def get_groups(domain_controller: DomainController, distinguishedName: str, nested: Optional[bool] = False) -> List[Group]:
    """Get groups from a group's membership list.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __distinguishedName:__ the distinguished name of the group

    - __nested:__ set to `true` to include nested group members in the search

??? success "Output"
    A list of [Group Dictionaries](../inputs_outputs/#group).

??? warning
    Getting nested group memberships can be slow.

??? example
    Getting all groups in _Example Group_, & saving them as a new variable called _group_list_.
    ```YAML linenums="1"
    - active_directory.groups.get_groups:
        distinguishedName: "CN=Example Group,CN=Users,DC=Example,DC=Domain"
        nested: true
      load:
        domain_controller: domain_controller_info
      save: group_list
    ```
    """
    return [
        GroupData(**group.__dict__).__dict__
        for group in list(
            filter(
                lambda record: isinstance(record, groups.Group),
                groups.members(domain_controller, distinguished_name=distinguishedName, nested=nested),
            )
        )
    ]


@ldap_function
def get_users(domain_controller: DomainController, distinguishedName: str, nested: Optional[bool] = False) -> List[User]:
    """Get users from a group's membership list.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __distinguishedName:__ the distinguished name of the group

    - __nested:__ set to `true` to include nested group members in the search

??? success "Output"

    A list of [User Dictionaries](../inputs_outputs/#user).

??? warning
    Getting nested group memberships can be slow.

??? example
    Getting all users in _Example Group_, & saving them as a new variable called _user_list_.
    ```YAML linenums="1"
    - active_directory.groups.get_users:
        distinguishedName: "CN=Example Group,CN=Users,DC=Example,DC=Domain"
        nested: true
      load:
        domain_controller: domain_controller_info
      save: user_list
    ```
    """
    return [
        UserData(**user.__dict__).__dict__
        for user in list(
            filter(
                lambda record: isinstance(record, users.User),
                groups.members(domain_controller, distinguished_name=distinguishedName, nested=nested),
            )
        )
    ]


@ldap_function
def get_computers(domain_controller: DomainController, distinguishedName: str, nested: Optional[bool] = False) -> List[Computer]:
    """Get computers from a group's membership list.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __distinguishedName:__ the distinguished name of the group

    - __nested:__ set to `true` to include nested group members in the search

??? success "Output"
    A list of [Computer Dictionaries](../inputs_outputs/#group).

??? warning
    Getting nested group memberships can be slow.

??? example
    - Getting all computers in _Example Group_

    - The results are saved as a new variable called `computer_list`

    ```YAML linenums="1"
    - active_directory.groups.get_computers:
        distinguishedName: "CN=Example Group,CN=Users,DC=Example,DC=Domain"
        nested: true
      load:
        domain_controller: domain_controller_info
      save: computer_list
    ```
    """
    return [
        ComputerData(**computer.__dict__).__dict__
        for computer in list(
            filter(
                lambda record: isinstance(record, computers.Computer),
                groups.members(domain_controller, distinguished_name=distinguishedName, nested=nested),
            )
        )
    ]


@ldap_function
def search(domain_controller: DomainController, search_params: Dict[str, Any], search_base: Optional[str] = None) -> List[Group]:
    """Search for groups using LDAP attributes & values.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __search_params:__ a dictionary containing [group keys](../inputs_outputs#group) & values to use in the search

    - __search_base:__ a distinguishedName to use as the root of the search (defaults to the root of the domain)

??? success "Output"
    A list of [Group Dictionaries](../inputs_outputs/#group).

??? example

    Searching for all groups in the __Users__ CN whose `cn` starts with __admin*__:

    ```YAML linenums="1"
    - active_directory.groups.search:
        search_params:
          cn: admin*
        search_base: CN=Users,DC=Example,DC=Domain,DC=Com
      load:
        domain_controller: domain_controller_info
      save: groups
    ```

!!! tip

    You can use `*` as a wildcard at the end of search values.
"""
    return [
        GroupData(**group.__dict__).__dict__
        for group in groups.search(domain_controller, search_base=search_base, **search_params)
    ]


@ldap_function
@idempotent_create
def add_member(domain_controller: DomainController, group_distinguishedName: str, member_distinguishedName: str) -> None:
    """Add a member to a group.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __group_distinguishedName:__ the distinguishedName of the group

    - __member_distinguishedName:__ the distinguishedName of the new member

??? success "Output"
    Nothing is outputted by this action.

??? example
    Adding a user to a group.

    - The user & group are searched for using the _sAMAccountName_ attribute

    - They are saved as new variables `user` & `group`

    - The required _distinguishedName_ values are supplied using fields from the `user` & `group` variables

    ```YAML linenums="1"
    - active_directory.groups.search:
        sAMAccountName: example.group
      load:
        domain_controller: domain_controller_info
      save: group

    - active_directory.users.search:
        sAMAccountName: example.user
      load:
        domain_controller: domain_controller_info
      save: user

    - active_directory.groups.add_member:
      load:
        group_distinguishedName: group.distinguishedName
        member_distinguishedName: member.distinguishedName
        domain_controller: domain_controller_info
    ```
"""
    return groups.add_member(
        domain_controller, distinguished_name=group_distinguishedName, member_distinguished_name=member_distinguishedName,
    )


@ldap_function
@idempotent_remove
def remove_member(domain_controller: DomainController, group_distinguishedName: str, member_distinguishedName: str) -> None:
    """Remove a member from a group.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __group_distinguishedName:__ the _distinguishedName_ of the group

    - __member_distinguishedName:__ the _distinguishedName_ of the member to remove

??? success "Output"
    Nothing is outputted by this action.

??? example

    Removing user _Example User_ from the group _Example Group_.

    ```YAML linenums="1"
    - active_directory.groups.remove_member:
        group_distinguishedName: "CN=Example Group,CN=Users,DC=Example,DC=Domain"
        member_distinguishedName: "CN=Example User,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
    """
    return groups.remove_member(
        domain_controller, distinguished_name=group_distinguishedName, member_distinguished_name=member_distinguishedName,
    )


@ldap_function
@idempotent_create
def create(
    domain_controller: DomainController,
    sAMAccountName: str,
    distinguishedName: str,
    extra_params: Optional[Dict[str, Any]] = None,
) -> None:
    """Create a new security group.

!!! question "Minimum Plugin Version: 1.1.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __sAMAccountName:__ the new group sAMAccountName

    - __distinguishedName:__ the new group distinguishedName

    - __extra_params:__ a dictionary containing any extra LDAP attributes & values for the group

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.groups.create:
        sAMAccountName: example.group
        distinguishedName: "CN=Example Group,CN=Users,DC=Example,DC=Domain"
        extra_params:
          info: An example Security Group
          cn: Example Group
      load:
        domain_controller: domain_controller_info
    ```
    """
    return groups.create(
        domain_controller, distinguished_name=distinguishedName, sAMAccountName=sAMAccountName, **extra_params or {}
    )


@ldap_function
@idempotent_remove
def delete(domain_controller: DomainController, distinguishedName: str) -> None:
    """Delete a group.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the group to delete

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.groups.delete:
        distinguishedName: "CN=Example Group,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
    """
    logger.info(f"deleting group with distinguished name {distinguishedName}")
    return groups.delete(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def move(domain_controller: DomainController, distinguishedName: str, parent_distinguishedName: str) -> None:
    """Move a group to a different OU or CN.

!!! question "Minimum Plugin Version: 1.5.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the group to move

    - __parent_distinguishedName:__ the distinguishedName of the OU or CN to move the group to

??? success "Output"

    Nothing is outputted by this action.

??? example

    Moving a group from the 'Users' CN to the 'Staff Groups' OU:

    ```YAML linenums="1"
    - active_directory.groups.move:
        distinguishedName: "CN=Example Group,CN=Users,DC=Example,DC=Domain"
        parent_distinguishedName: "OU=Staff Groups,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
    """
    return groups.move(domain_controller, distinguished_name=distinguishedName, new_parent=parent_distinguishedName)


@ldap_function
def delete_attribute(domain_controller: DomainController, distinguishedName: str, name: str, value: Union[str, int]) -> None:
    """Delete a certain value from a group attribute.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the group

    - __name:__ the name of the LDAP attribute

    - __value:__ the value to delete

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.groups.get_interactive:
        cn: Example Group
      load:
        domain_controller: domain_controller_info
      save: group

    - active_directory.groups.delete_attribute:
        name: info
      load:
        distinguishedName: group.distinguishedName
        value: group.info
        domain_controller: domain_controller_info
    ```
    """
    return groups.delete_attribute(domain_controller, distinguished_name=distinguishedName, name=name, value=value)


@ldap_function
def set_attribute(domain_controller: DomainController, distinguishedName: str, name: str, value: Union[str, int]) -> None:
    """Set a group attribute value.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the group

    - __name:__ the name of the LDAP attribute

    - __value:__ the value to set

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.groups.get_interactive:
        cn: Example Group
      load:
        domain_controller: domain_controller_info
      save: group

    - active_directory.groups.set_attribute:
        name: info
      load:
        distinguishedName: group.distinguishedName
        value: Example Info Value
        domain_controller: domain_controller_info
    ```
    """
    return groups.set_attribute(domain_controller, distinguished_name=distinguishedName, name=name, value=value)


@ldap_function
def replace_attribute(domain_controller: DomainController, distinguishedName: str, name: str, value: Union[str, int]) -> None:
    """Replace a group attribute value.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"
    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the group

    - __name:__ the name of the LDAP attribute

    - __value:__ the value to set

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.groups.get_interactive:
        cn: Example Group
      load:
        domain_controller: domain_controller_info
      save: group

    - active_directory.groups.replace_attribute:
        name: info
        value: Example Info Value
      load:
        distinguishedName: group.distinguishedName
        domain_controller: domain_controller_info
    ```
    """
    return groups.replace_attribute(domain_controller, distinguished_name=distinguishedName, name=name, value=value)


def output_table(text: str, group_list: List[Group]) -> None:
    """Display a list of groups in a table.

The table will have the following columns:

- Common Name
- Group Type
- Distinguished Name

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"
    - __text:__ the title of the table

    - __group_list:__ a list of [Group Dictionaries](../inputs_outputs/#group)

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.groups.search:
        cn: Operations
      load:
        domain_controller: domain_controller_info
      save: operations_groups

    - active_directory.groups.output_table:
        text: Operations Groups
      load:
        group_list: operations_groups
    ```
    """
    ui.output_table(
        text,
        header=["Common Name", "Type", "Distinguished Name"],
        data=[[group.get(field) or "" for field in ["cn", "groupType", "distinguishedName"]] for group in group_list],
    )


def output_custom_table(text: str, header: List[str], attributes: List[str], group_list: List[Group]) -> None:
    """Display a list of groups in a table using custom headers & LDAP attributes.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __header:__ a list of column names

    - __attributes:__ a list of LDAP attributes (one for each column name in the `header`)

    - __user_list:__ a list of [Group Dictionaries](../inputs_outputs/#group) to display in the table

??? success "Output"
    Nothing is outputted by this action.

??? example
    Showing `cn` & _groupType_ for groups in the variable `test_groups`.

    ```YAML linenums="1"
    - active_directory.groups.output_custom_table:
        text: Custom Groups Table
        header:
          - Common Name
          - Group Type
        attributes:
          - cn
          - groupType
      load:
        group_list: test_groups
    ```
    """
    ui.output_table(text, header=header, data=[[group.get(attr) or "" for attr in attributes] for group in group_list])


def input_table(text: str, group_list: List[Group], minimum: Optional[int] = None, maximum: Optional[int] = None) -> List[Group]:
    """Display a list of groups in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Common Name
- Type
- Distinguished Name

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ The title of the table.

    - __group_list:__ A list of [Group Dictionaries](../inputs_outputs/#group) to display in the table

    - __minimum:__ The minimum number of acceptable selections.

    - __maximum:__ The maximum number of acceptable selections.

??? success "Output"
    A list of [Group Dictionaries](../inputs_outputs/#group).

??? tip

    - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.
    - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number
        of selections.

??? example
    - Getting all groups whose common names start with _Test_

    - Saving the results as a new variable called `test_groups`

    - Using this action to show the `test_groups` in a table, requiring the task operator selects at least 1

    - The selection is saved as a new variable called `selected_groups`

    ```YAML linenums="1"
    - active_directory.groups.search:
        cn: Test*
      load:
        domain_controller: domain_controller_info
      save: test_groups

    - active_directory.groups.input_table:
        text: "Please Select >= 1 Group(s)"
        minimum: 1
      load:
        group_list: test_groups
      save: selected_groups
    ```
    """
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    while True:
        indices = ui.input_table(
            text,
            header=["Common Name", "Type", "Distinguished Name"],
            data=[[group.get(field) or "" for field in ["cn", "groupType", "distinguishedName"]] for group in group_list],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [group_list[i] for i in indices]


@ldap_function
def exists(domain_controller: DomainController, **search_params: Union[str, int, None]) -> bool:
    """Search for groups using LDAP attributes & values to identify if any were found.

!!! question "Minimum Plugin Version: 1.0.0"


??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __search_params:__ a dictionary containing [group keys](../inputs_outputs#group) & values to use in the search


??? success "Output"

    A boolean is outputted by this action

    - `true` if one or more groups are found

    - `false` if no groups are found


??? example

    Searching using a unique attribute:

    ```YAML linenums="1"
    - active_directory.groups.exists:
        distinguishedName: "CN=Group Name,OU=Example,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
      save: item_exists
    ```

    Searching using a non-unique attribute:

    ```YAML linenums="1"
    - active_directory.groups.exists:
        cn: "My Group*"
      load:
        domain_controller: domain_controller_info
      save: item_exists
    ```

!!! tip "Wildcard Searching"

    It is possible to use `*` as a wildcard at the __end__ of search values, but this can make the search slow.

    """
    return bool(groups.search(domain_controller, **search_params))
