import logging

from datacaster.classes import CastDataClass

from typing import TypedDict, Optional, List

logger = logging.getLogger(__name__)


DateString = Optional[str]


def _format_date_string(value):
    return value.date_object.strftime("%H:%M:%S %d/%m/%Y") if value else None


def _get_decoded_ldap_string(value):
    return value.string if hasattr(value, "string") else None


class DomainController(TypedDict):
    """
- Contains Domain Controller connection details

- It is used as an _input_ to many of the actions in this plugin

- The example below shows the dictionary structure in YAML

- See the ___required___ and _optional_ keys underneath the example

!!! warning "Supplying Credentials"

    You should ___always___ use a PPA Vault integration to provide credentials to a plugin action.

!!! example

    ```yaml linenums="1"
    domain_controller:
      address: 1.2.3.4
      domain: example.domain.net
      port: 636
      username: [username]
      password: [password]
    ```

!!! danger "Required Keys"

    __address__: Domain Controller IP or DNS address.

    __domain__: FQDN of the Active Directory domain.

    __username__: Username for authentication.

    __password__: Password for authentication.

!!! info "Optional Keys"

    __port__: The LDAPS port on the Domain Controller (defaults to `636`).

    """

    address: str
    port: Optional[int]
    domain: str
    username: str
    password: str


class User(TypedDict):
    """
- Contains Active Directory User LDAP attributes

- Expand below to see each key and a link to the relevant Microsoft article

??? info "All User Keys"

    __accountExpires__: Date string or null [[Account-Expires](https://docs.microsoft.com/en-us/windows/win32/adschema/a-accountexpires)].

    __badPasswordTime__: Date string or null [[Bad-Password-Time](https://docs.microsoft.com/en-us/windows/win32/adschema/a-badpasswordtime)].

    __badPwdCount__: Integer [[Bad-Pwd-Count](https://docs.microsoft.com/en-us/windows/win32/adschema/a-badpwdcount)].

    __carLicense__: String or null [[carLicense](https://docs.microsoft.com/en-us/windows/win32/adschema/a-carlicense)].

    __cn__: String or null [[Common-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-cn)].

    __countryCode__: String or null [[Country-Code](https://docs.microsoft.com/en-us/windows/win32/adschema/a-countrycode)].

    __displayName__: String or null [[Display-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-displayname)].

    __distinguishedName__: String [[Obj-Dist-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-distinguishedname)].

    __givenName__: String or null [[Given-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-givenname)].

    __info__: String or null [[Comment](https://docs.microsoft.com/en-us/windows/win32/adschema/a-info)].

    __lastLogoff__: Date string or null [[Last-Logoff](https://docs.microsoft.com/en-us/windows/win32/adschema/a-lastlogoff)].

    __lastLogon__: Date string or null [[Last-Logon](https://docs.microsoft.com/en-us/windows/win32/adschema/a-lastlogon)].

    __lastLogonTimestamp__: Date string or null [[Last-Logon-Timestamp](https://docs.microsoft.com/en-us/windows/win32/adschema/a-lastlogontimestamp)].

    __lockoutTime__: Date string or null [[Lockout-Time](https://docs.microsoft.com/en-us/windows/win32/adschema/a-lockouttime)].

    __logonCount__: Integer [[Logon-Count](https://docs.microsoft.com/en-us/windows/win32/adschema/a-logoncount)].

    __mail__: String or null [[E-mail-Addresses](https://docs.microsoft.com/en-us/windows/win32/adschema/a-mail)].

    __manager__: String or null [[Manager](https://docs.microsoft.com/en-us/windows/win32/adschema/a-manager)].

    __memberOf__: List of [Group Dictionaries](#group) [[Is-Member-Of-DL](https://docs.microsoft.com/en-us/windows/win32/adschema/a-memberof)].

    __name__: String or null [[RDN](https://docs.microsoft.com/en-us/windows/win32/adschema/a-name)].

    __objectGUID__: String [[Object-Guid](https://docs.microsoft.com/en-us/windows/win32/adschema/a-objectguid)].

    __objectSid__: String [[Object-Sid](https://docs.microsoft.com/en-us/windows/win32/adschema/a-objectsid)].

    __pwdLastSet__: Date string or null [[Pwd-Last-Set](https://docs.microsoft.com/en-us/windows/win32/adschema/a-pwdlastset)].

    __sAMAccountName__: String [[Sam-Account-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-samaccountname)].

    __sAMAccountType__: Integer [[Sam-Account-Type](https://docs.microsoft.com/en-us/windows/win32/adschema/a-samaccounttype)].

    __sn__: String or null [[Surname](https://docs.microsoft.com/en-us/windows/win32/adschema/a-sn)].

    __telephoneNumber__: String or null [[Telephone-Number](https://docs.microsoft.com/en-us/windows/win32/adschema/a-telephonenumber)].

    __thumbnailPhoto__: String or null [[Picture](https://docs.microsoft.com/en-us/windows/win32/adschema/a-thumbnailphoto)].

    __userAccountControl__: Integer [[User-Account-Control](https://docs.microsoft.com/en-us/windows/win32/adschema/a-useraccountcontrol)].

    __userPrincipalName__: String or null [[User-Principal-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-userprincipalname)].

!!! tip "See more about Users [here](https://docs.microsoft.com/en-gb/windows/win32/adschema/c-user)"

    """

    accountExpires: DateString
    badPasswordTime: DateString
    badPwdCount: Optional[int]
    carLicense: Optional[List[str]]
    cn: Optional[str]
    countryCode: Optional[str]
    displayName: Optional[str]
    distinguishedName: str
    givenName: Optional[str]
    info: Optional[str]
    lastLogoff: DateString
    lastLogon: DateString
    lastLogonTimestamp: DateString
    lockoutTime: DateString
    logonCount: Optional[int]
    mail: Optional[str]
    manager: Optional[str]
    memberOf: Optional[List[str]]
    name: Optional[str]
    objectGUID: str
    objectSid: str
    pwdLastSet: DateString
    sAMAccountName: str
    sAMAccountType: Optional[int]
    sn: Optional[str]
    telephoneNumber: Optional[str]
    thumbnailPhoto: Optional[str]
    userAccountControl: int
    userPrincipalName: Optional[str]


class UserData(CastDataClass):
    __annotations__ = User.__dict__["__annotations__"]
    __class_config__ = {
        "cast_functions": {
            "types": {DateString: _format_date_string},
            "fields": {
                "objectGUID": _get_decoded_ldap_string,
                "objectSid": _get_decoded_ldap_string,
                "thumbnailPhoto": _get_decoded_ldap_string,
            },
        },
        "always_cast": ["objectGUID", "objectSid", "thumbnailPhoto"],
    }


class Group(TypedDict):
    """
- Contains Active Directory Group LDAP attributes

- Expand below to see each key and a link to the relevant Microsoft article

??? info "All Group Keys"

    __cn__: String or null [[Common-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-cn)].

    __description__: String or null [[Description](https://docs.microsoft.com/en-us/windows/win32/adschema/a-description)].

    __distinguishedName__: String [[Obj-Dist-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-distinguishedname)].

    __groupType__: String [[Group-Type](https://docs.microsoft.com/en-gb/windows/win32/adschema/a-grouptype)].

    __info__: String or null [[Comment](https://docs.microsoft.com/en-us/windows/win32/adschema/a-info)].

    __managedBy__: String or null [[ManagedBy](https://docs.microsoft.com/en-us/windows/win32/adschema/a-managedby)].

    __member__: List of [User](#user), [Group](#group), or [Computer](#computer) dictionaries [[Member](https://docs.microsoft.com/en-us/windows/win32/adschema/a-member)].

    __name__: String or null [[RDN](https://docs.microsoft.com/en-us/windows/win32/adschema/a-name)].

    __objectGUID__: String [[Object-Guid](https://docs.microsoft.com/en-us/windows/win32/adschema/a-objectguid)].

    __objectSid__: String [[Object-Sid](https://docs.microsoft.com/en-us/windows/win32/adschema/a-objectsid)].

    __sAMAccountName__: String [[Sam-Account-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-samaccountname)].

    __sAMAccountType__: Integer [[Sam-Account-Type](https://docs.microsoft.com/en-us/windows/win32/adschema/a-samaccounttype)].

!!! tip "See more about Groups [here](https://docs.microsoft.com/en-gb/windows/win32/adschema/c-group)"

    """

    cn: Optional[str]
    description: Optional[str]
    distinguishedName: str
    groupType: str
    info: Optional[str]
    managedBy: Optional[str]
    member: Optional[List[str]]
    name: Optional[str]
    objectGUID: str
    objectSid: str
    sAMAccountName: str
    sAMAccountType: int


class GroupData(CastDataClass):
    __annotations__ = Group.__dict__["__annotations__"]
    __class_config__ = {
        "cast_functions": {
            "types": {DateString: _format_date_string},
            "fields": {"objectGUID": _get_decoded_ldap_string, "objectSid": _get_decoded_ldap_string},
        },
        "always_cast": ["objectGUID", "objectSid"],
    }


class Computer(TypedDict):
    """
- Contains Active Directory Computer LDAP attributes

- Expand below to see each key and a link to the relevant Microsoft article

??? info "All Computer Keys"

    __accountExpires__: Date string or null [[Account-Expires](https://docs.microsoft.com/en-us/windows/win32/adschema/a-accountexpires)].

    __carLicense__: String or null [[carLicense](https://docs.microsoft.com/en-us/windows/win32/adschema/a-carlicense)].

    __cn__: String or null [[Common-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-cn)].

    __description__: String or null [[Description](https://docs.microsoft.com/en-us/windows/win32/adschema/a-description)].

    __displayName__: String or null [[Display-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-displayname)].

    __distinguishedName__: String [[Obj-Dist-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-distinguishedname)].

    __info__: String or null [[Comment](https://docs.microsoft.com/en-us/windows/win32/adschema/a-info)].

    __lastLogoff__: Date string or null [[Last-Logoff](https://docs.microsoft.com/en-us/windows/win32/adschema/a-lastlogoff)].

    __lastLogon__: Date string or null [[Last-Logon](https://docs.microsoft.com/en-us/windows/win32/adschema/a-lastlogon)].

    __lastLogonTimestamp__: Date string or null [[Last-Logon-Timestamp](https://docs.microsoft.com/en-us/windows/win32/adschema/a-lastlogontimestamp)].

    __lockoutTime__: Date string or null [[Lockout-Time](https://docs.microsoft.com/en-us/windows/win32/adschema/a-lockouttime)].

    __logonCount__: Integer [[Logon-Count](https://docs.microsoft.com/en-us/windows/win32/adschema/a-logoncount)].

    __mail__: String or null [[E-mail-Addresses](https://docs.microsoft.com/en-us/windows/win32/adschema/a-mail)].

    __mail__: String or null [[Managed-By](https://docs.microsoft.com/en-us/windows/win32/adschema/a-managedby)].

    __memberOf__: List of [Group Dictionaries](#group) [[Is-Member-Of-DL](https://docs.microsoft.com/en-us/windows/win32/adschema/a-memberof)].

    __name__: String or null [[RDN](https://docs.microsoft.com/en-us/windows/win32/adschema/a-name)].

    __objectGUID__: String [[Object-Guid](https://docs.microsoft.com/en-us/windows/win32/adschema/a-objectguid)].

    __objectSid__: String [[Object-Sid](https://docs.microsoft.com/en-us/windows/win32/adschema/a-objectsid)].

    __operatingSystem__: String or null [[Operating-System](https://docs.microsoft.com/en-us/windows/win32/adschema/a-operatingsystem)].

    __operatingSystemHotfix__: String or null [[Operating-System-Hotfix](https://docs.microsoft.com/en-us/windows/win32/adschema/a-operatingsystemhotfix)].

    __operatingSystemServicePack__: String or null [[Operating-System-Service-Pack](https://docs.microsoft.com/en-us/windows/win32/adschema/a-operatingsystemservicepack)].

    __operatingSystemVersion__: String or null [[Operating-System-Version](https://docs.microsoft.com/en-us/windows/win32/adschema/a-operatingsystemversion)].

    __pwdLastSet__: Date string or null [[Pwd-Last-Set](https://docs.microsoft.com/en-us/windows/win32/adschema/a-pwdlastset)].

    __sAMAccountName__: String [[Sam-Account-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-samaccountname)].

    __sAMAccountType__: Integer [[Sam-Account-Type](https://docs.microsoft.com/en-us/windows/win32/adschema/a-samaccounttype)].

    __userAccountControl__: Integer [[User-Account-Control](https://docs.microsoft.com/en-us/windows/win32/adschema/a-useraccountcontrol)].

    __userPrincipalName__: String or null [[User-Principal-Name](https://docs.microsoft.com/en-us/windows/win32/adschema/a-userprincipalname)].

!!! tip "See more about Computers [here](https://docs.microsoft.com/en-gb/windows/win32/adschema/c-computer)"
    """

    accountExpires: DateString
    carLicense: Optional[List[str]]
    cn: Optional[str]
    description: Optional[str]
    displayName: Optional[str]
    distinguishedName: str
    info: Optional[str]
    lastLogoff: Optional[str]
    lastLogon: DateString
    lastLogonTimestamp: DateString
    lockoutTime: DateString
    logonCount: Optional[int]
    mail: Optional[str]
    managedBy: Optional[str]
    memberOf: Optional[List[str]]
    name: Optional[str]
    objectGUID: str
    objectSid: str
    operatingSystem: Optional[str]
    operatingSystemHotfix: Optional[str]
    operatingSystemServicePack: Optional[str]
    operatingSystemVersion: Optional[str]
    pwdLastSet: DateString
    sAMAccountName: str
    sAMAccountType: int
    userAccountControl: int
    userPrincipalName: Optional[str]


class ComputerData(CastDataClass):
    __annotations__ = Computer.__dict__["__annotations__"]
    __class_config__ = {
        "cast_functions": {
            "types": {DateString: _format_date_string},
            "fields": {"objectGUID": _get_decoded_ldap_string, "objectSid": _get_decoded_ldap_string},
        },
        "always_cast": ["objectGUID", "objectSid"],
    }
